from copy import deepcopy
import logging
import numpy as np
import os.path as osp
import strax
from strax import exporter
import straxen


NT_AUX_INSTALLED = False
STRAX_AUX_INSTALLED = False

try:
    import ntauxfiles
    NT_AUX_INSTALLED = True
except (ModuleNotFoundError, ImportError):
    pass
try:
    import straxauxfiles
    STRAX_AUX_INSTALLED = True
except (ModuleNotFoundError, ImportError):
    pass


export, __all__ = exporter()
logging.basicConfig(handlers=[logging.StreamHandler()])
log = logging.getLogger('wfsim.resource')
log.setLevel('WARNING')


_cached_configs = dict()


@export
def load_config(config):
    """Create a Resource instance from the configuration

    Uses a cache to avoid re-creating instances from the same config
    """
    h = strax.deterministic_hash(Resource.config_to_file(config))
    if h in _cached_configs:
        return _cached_configs[h]
    result = Resource(config)
    _cached_configs[h] = result
    log.debug(f'Caching config file set {h}')
    return result


@export
class Resource:
    """
    Get the configs needed for running WFSim. Configs can be obtained in
        two ways:
        1. Get it directly from the mongo database. This only needs the
            name of the file.
        2. Load it with straxen get_resource, this can either:
            Download from a public repository
            Read from local cache
            Download from a private repository if credentials 
            are properly setup
    """
    @staticmethod
    def config_to_file(config):
        """
        Find and complete all file paths

        returns dictionary mapping item name to path
        """
        if config is None:
            config = dict()

        files = {}
        if config['detector'] == 'XENON1T':
            files.update({
                'photon_area_distribution': 'XENON1T_spe_distributions.csv',
                's1_lce_correction_map': 'XENON1T_s1_xyz_ly_kr83m_SR1_pax-680_fdc-3d_v0.json',
                's1_pattern_map': 'XENON1T_s1_xyz_patterns_interp_corrected_MCv2.1.0.json.gz',
                's2_correction_map': 'XENON1T_s2_xy_ly_SR1_v2.2.json',
                's2_pattern_map': 'XENON1T_s2_xy_patterns_top_corrected_MCv2.1.0.json.gz',
                'photon_ap_cdfs': 'x1t_pmt_afterpulse_config.pkl.gz',
                'fdc_3d': 'XENON1T_FDC_SR1_data_driven_time_dependent_3d_correction_tf_nn_part1_v1.json.gz',
                'ele_ap_pdfs': 'x1t_se_afterpulse_delaytime.pkl.gz',
                'noise_file': 'x1t_noise_170203_0850_00_small.npz'
            })

        elif config['detector'] == 'XENONnT':
            files.update({
                'photon_area_distribution': 'XENONnT_spe_distributions_20210305.csv',
                's1_pattern_map': 'XENONnT_s1_xyz_patterns_LCE_corrected_qes_MCva43fa9b_wires.pkl',
                's1_lce_correction_map': 'XENONnT_s1_xyz_LCE_corrected_qes_MCva43fa9b_wires.json.gz',
                's2_pattern_map': 'XENONnT_s2_xy_patterns_LCE_corrected_qes_MCva43fa9b_wires.pkl',
                's2_correction_map': 'XENONnT_s2_xy_map_v4_210503_mlp_3_in_1_iterated.json',
                'se_gain_map': 'XENONnT_se_xy_map_v1_mlp.json',
                'photon_ap_cdfs': 'XENONnT_pmt_afterpulse_config_012605.json.gz',
                's2_luminescence': 'XENONnT_GARFIELD_B1d5n_C30n_G1n_A6d5p_T1d5n_PMTs1d5n_FSR0d95n.npz',
                "s2_luminescence_gg": "garfield_timing_map_gas_gap_sr0.npy",
                'gas_gap_map': 'gas_gap_warping_map_January_2021.pkl',
                'garfield_gas_gap_map': 'garfield_gas_gap_map_sr0.json',
                'ele_ap_pdfs': 'x1t_se_afterpulse_delaytime.pkl.gz',
                'noise_file': 'x1t_noise_170203_0850_00_small.npz',
                'fdc_3d': 'XnT_3D_FDC_xyt_dummy_all_zeros_v0.1.json.gz',
                'field_dependencies_map': '',
                's1_time_spline': 'XENONnT_s1_proponly_va43fa9b_wires_20200625.json.gz',
                's2_time_spline': '',
             })

        elif config['detector'] == 'XENONnT_neutron_veto':
            files.update({
                'photon_area_distribution': 'XENONnT_spe_distributions_nveto_013071.csv',
                'nv_pmt_qe': 'nveto_pmt_qe.json',
                'noise_file': 'xnt_noise_nveto_014901.npz'
            })

        else:
            raise ValueError(f"Unsupported detector {config['detector']}")

        # Allowing user to replace default with specified files
        for k in set(config).intersection(files):
            files[k] = config[k]

        commit = 'master'  # Replace this by a commit hash if you feel solid and responsible
        if config.get('url_base', False):
            files['url_base'] = config['url_base']
        elif config['detector'] == "XENON1T":
            files['url_base'] = f'https://raw.githubusercontent.com/XENONnT/strax_auxiliary_files/{commit}/sim_files'
        else:
            files['url_base'] = f'https://raw.githubusercontent.com/XENONnT/private_nt_aux_files/{commit}/sim_files'

        return files

    @staticmethod
    def get_file_path(base, fname):
        """Find the full path to the resource file
        Try 4 methods in the following order
        1. The base is not url, return base + name
        2. If ntauxfiles (straxauxfiles) is installed, return will be package dir + name
           pip install won't work, try python setup.py in the packages
        3. Download the latest version using straxen mongo downloader from database,
           return the cached file path + md5
        4. Download using straxen get_resource from the url (github raw)
           simply return base + name
            Be careful with the forth options, straxen creates
            cache files that might not be updated with the latest github commit.

        """
        if not fname:
            log.warning(f"A file has value False, assuming this is intentional.")
            return

        if fname.startswith('/'):
            log.warning(f"Using local file {fname} for a resource. "
                        f"Do not set this as a default or TravisCI tests will break")
            return fname
        
        if base.startswith('/'):
            log.warning(f"Using local folder {base} for all resources. "
                        f"Do not set this as a default or TravisCI tests will break")
            return osp.join(base, fname)

        if NT_AUX_INSTALLED:
            # You might want to use this, for example if you are a developer
            if fname in ntauxfiles.list_private_files():
                log.warning(f"Using the private repo to load {fname} locally")
                fpath = ntauxfiles.get_abspath(fname)
                log.info(f"Loading {fname} is successfully from {fpath}")
                return fpath

        if STRAX_AUX_INSTALLED:
            if fname in straxauxfiles.list_aux_files():
                log.warning(f"Using the public repo to load {fname} locally")
                fpath = straxauxfiles.get_abspath(fname)
                log.info(f"Loading {fname} is successfully from {fpath}")
                return fpath

        try:
            # https://straxen.readthedocs.io/en/latest/config_storage.html
            # downloading-xenonnt-files-from-the-database  # noqa

            # we need to add the straxen.MongoDownloader() in this
            # try: except NameError: logic because the NameError
            # gets raised if we don't have access to utilix.
            downloader = straxen.MongoDownloader()
            # FileNotFoundError, ValueErrors can be raised if we
            # cannot load the requested config
            fpath = downloader.download_single(fname)
            log.warning(f"Loading {fname} from mongo downloader to {fpath}")
            return fname  # Keep the name and let get_resource do its thing

        except (FileNotFoundError, ValueError, NameError, AttributeError):
            log.info(f"Mongo downloader not possible or does not have {fname}")

        # We cannot download the file from the database. We need to
        # try to get a placeholder file from a URL.
        furl = osp.join(base, fname)
        log.warning(f'{fname} did not download, trying download from {base}')
        return furl

    def __init__(self, config=None):
        files = self.config_to_file(config)
        log.debug('Getting\n' + '\n'.join([f'{k}: {v}' for k, v in files.items()]))

        for k, v in files.items():
            if isinstance(v, list):
                # It's a dummy map call, do nothing
                continue
            if k == 'url_base':
                continue
            if v == '':
                log.warning(f'{k} has no path so this config file is set to None')
                files[k] = None
                continue

            log.debug(f'Obtaining {k} from {v}')
            files[k] = self.get_file_path(files['url_base'], v)

        if config.get('detector', 'XENONnT') == 'XENON1T':
            self.s1_pattern_map = make_map(files['s1_pattern_map'], fmt='json.gz')
            self.s1_lce_correction_map = make_map(files['s1_lce_correction_map'], fmt='json')
            self.s2_correction_map = make_map(files['s2_correction_map'], fmt='json')
            self.s2_pattern_map = make_map(files['s2_pattern_map'], fmt='json.gz')
            self.fdc_3d = make_map(files['fdc_3d'], fmt='json.gz')

            # Gas gap warping map
            if config.get('enable_gas_gap_warping', False):
                self.gas_gap_length = make_map(["constant dummy", 0.25, [254, ]])

            # Photon After Pulses
            if config.get('enable_pmt_afterpulses', False):
                self.uniform_to_pmt_ap = straxen.get_resource(files['photon_ap_cdfs'], fmt='pkl.gz')

            # Electron After Pulses
            if config.get('enable_electron_afterpulses', False):
                self.uniform_to_ele_ap = straxen.get_resource(files['ele_ap_pdfs'], fmt='pkl.gz')

        elif config.get('detector', 'XENONnT') == 'XENONnT':
            pmt_mask = np.array(config['gains']) > 0  # Converted from to pe (from cmt by default)
            self.s1_pattern_map = make_patternmap(files['s1_pattern_map'], fmt='pkl', pmt_mask=pmt_mask)
            self.s2_pattern_map = make_patternmap(files['s2_pattern_map'], fmt='pkl', pmt_mask=pmt_mask)
            self.se_gain_map = make_map(files['se_gain_map'])
#             self.s2_correction_map = make_map(files['s2_correction_map'], fmt = 'json')

            # if there is a (data driven!) map, load it. If not make it  from the pattern map
            if files['s1_lce_correction_map']:
                self.s1_lce_correction_map = make_map(files['s1_lce_correction_map'])
            else:
                lymap = deepcopy(self.s1_pattern_map)
                # AT: this scaling with mast is redundant to `make_patternmap`, but keep it in for now
                lymap.data['map'] = np.sum(lymap.data['map'][:][:][:], axis=3, keepdims=True, where=pmt_mask)
                lymap.__init__(lymap.data)
                self.s1_lce_correction_map = lymap
            # making S2 aft scaling (if provided)
            if 's2_mean_area_fraction_top' in config.keys():
                avg_s2aft_=config['s2_mean_area_fraction_top']
                if avg_s2aft_>=0.0:
                    if isinstance(files['s2_pattern_map'], list):
                        log.warning(f'Scaling of S2 AFT with dummy map, this will have no effect!')
                    else:
                        s2map=deepcopy(self.s2_pattern_map)
                        s2map_topeff_=s2map.data['map'][...,0:config['n_top_pmts']].sum(axis=2)
                        s2map_toteff_=s2map.data['map'].sum(axis=2)
                        orig_aft_=np.mean((s2map_topeff_/s2map_toteff_)[s2map_toteff_>0.0])
                        # getting scales for top/bottom separately to preserve total efficiency
                        scale_top_=avg_s2aft_/orig_aft_
                        scale_bot_=(1 - avg_s2aft_)/(1 - orig_aft_)
                        s2map.data['map'][:,:,0:config['n_top_pmts']]*=scale_top_
                        s2map.data['map'][:,:,config['n_top_pmts']:config['n_tpc_pmts']]*=scale_bot_
                        self.s2_pattern_map.__init__(s2map.data)

            # if there is a (data driven!) map, load it. If not make it  from the pattern map
            if files['s2_correction_map']:
                self.s2_correction_map = make_map(files['s2_correction_map'], fmt = 'json')
            else:
                s2cmap = deepcopy(self.s2_pattern_map)
                # Lower the LCE by removing contribution from dead PMTs
                # AT: masking is a bit redundant due to PMT mask application in make_patternmap
                s2cmap.data['map'] = np.sum(s2cmap.data['map'][:][:], axis=2, keepdims=True, where=pmt_mask)
                # Scale by median value
                s2cmap.data['map'] = s2cmap.data['map'] / np.median(s2cmap.data['map'][s2cmap.data['map'] > 0])
                s2cmap.__init__(s2cmap.data)
                self.s2_correction_map = s2cmap

            # Garfield luminescence timing samples
            # if config.get('s2_luminescence_model', False) == 'garfield':
            if 'garfield_gas_gap' in config.get('s2_luminescence_model', ''):
                #garfield_gas_gap option is using (x,y) -> gas gap (from the map) -> s2 luminescence
                #from garfield. This s2_luminescence_gg is indexed only by the gas gap, and
                #corresponds to electrons drawn directly below the anode
                
                s2_luminescence_map = straxen.get_resource(files['s2_luminescence_gg'], fmt='npy')
                self.s2_luminescence_gg = s2_luminescence_map
                self.garfield_gas_gap_map = make_map(files['garfield_gas_gap_map'], fmt = 'json')

            elif 'garfield' in config.get('s2_luminescence_model', ''):
                #This option indexes the luminescence times using the liquid level values
                #as well as the position between the full pitch of two gate wires
                gf_file_name = files['s2_luminescence']
                if gf_file_name.endswith('npy'):
                    s2_luminescence_map = straxen.get_resource(gf_file_name, fmt='npy')
                    self.s2_luminescence = s2_luminescence_map
                elif gf_file_name.endswith('npz'):
                    # Backwards compatibility from before #363 / #370
                    s2_luminescence_map = straxen.get_resource(gf_file_name, fmt='npy_pickle')['arr_0']
                    # Get directly the map for the simulated level
                    liquid_level_available = np.unique(s2_luminescence_map['ll'])  # available levels (cm)
                    liquid_level = config['gate_to_anode_distance'] - config['elr_gas_gap_length']  # cm
                    liquid_level = min(liquid_level_available, key=lambda x: abs(x - liquid_level))
                    self.s2_luminescence = s2_luminescence_map[s2_luminescence_map['ll'] == liquid_level]
                else:
                    raise ValueError(f'{gf_file_name} is of unknown format')

            if config.get('field_distortion_model', "none") == "inverse_fdc":
                self.fdc_3d = make_map(files['fdc_3d'], fmt='json.gz')
                self.fdc_3d.scale_coordinates([1., 1., - config['drift_velocity_liquid']])

            if config.get('field_distortion_model', "none") == "comsol":
                self.fd_comsol = make_map(config['field_distortion_comsol_map'], fmt='json.gz', method='RectBivariateSpline')

            # Gas gap warping map
            if config.get('enable_gas_gap_warping', False):
                gas_gap_map = straxen.get_resource(files['gas_gap_map'], fmt='pkl')
                self.gas_gap_length = lambda positions: gas_gap_map.lookup(*positions.T)

            # Field dependencies 
            # This config entry a dictionary of 5 items
            if any(config['enable_field_dependencies'].values()):
                field_dependencies_map = make_map(files['field_dependencies_map'], fmt='json.gz', method='RectBivariateSpline')
                self.drift_velocity_scaling=1.0
                # calculating drift velocity scaling to match total drift time for R=0 between cathode and gate
                if "norm_drift_velocity" in config['enable_field_dependencies'].keys():
                    if config['enable_field_dependencies']['norm_drift_velocity']:
                        norm_dvel = field_dependencies_map(np.array([ [0], [- config['tpc_length']]]).T,
                                                           map_name='drift_speed_map')[0]
                        norm_dvel*=1e-4
                        self.drift_velocity_scaling=config['drift_velocity_liquid']/norm_dvel
                def rz_map(z, xy, **kwargs):
                    r = np.sqrt(xy[:, 0]**2 + xy[:, 1]**2)
                    return field_dependencies_map(np.array([r, z]).T, **kwargs)
                self.field_dependencies_map = rz_map

            # Photon After Pulses
            if config.get('enable_pmt_afterpulses', False):
                self.uniform_to_pmt_ap = straxen.get_resource(files['photon_ap_cdfs'], fmt='json.gz')

            # S1 photon timing splines
            if config.get('s1_time_spline', False):
                self.s1_optical_propagation_spline = make_map(files['s1_time_spline'], fmt='json.gz',
                                                              method='RegularGridInterpolator')

            # Electron After Pulses
            if config.get('enable_electron_afterpulses', False):
                self.uniform_to_ele_ap = straxen.get_resource(files['ele_ap_pdfs'], fmt='pkl.gz')

            # S2 photons timing optical propagation delays
            if config.get('s2_time_spline', False):
                self.s2_optical_propagation_spline = make_map(files['s2_time_spline'])

        elif config.get('detector', 'XENONnT') == 'XENONnT_neutron_veto':
            # Neutron veto PMT QE as function of wavelength
                self.nv_pmt_qe = straxen.get_resource(files['nv_pmt_qe'], fmt='json')

        # SPE area distributions
        self.photon_area_distribution = straxen.get_resource(files['photon_area_distribution'], fmt='csv')

        # Noise sample
        if config.get('enable_noise', False):
            self.noise_data = straxen.get_resource(files['noise_file'], fmt='npy')['arr_0']
            n_channels = len(self.noise_data[0])
            log.warning(f'Using noise data {files["noise_file"]} with {n_channels} channels for {config["detector"]}')

        log.debug(f'{self.__class__.__name__} fully initialized')


def make_map(map_file, fmt=None, method='WeightedNearestNeighbors'):
    """Fetch and make an instance of InterpolatingMap based on map_file
    Alternatively map_file can be a list of ["constant dummy", constant: int, shape: list]
    return an instance of  DummyMap"""

    if isinstance(map_file, list):
        assert map_file[0] == 'constant dummy', ('Alternative file input can only be '
                                                 '("constant dummy", constant: int, shape: list')
        return DummyMap(map_file[1], map_file[2])

    elif isinstance(map_file, str):
        if fmt is None:
            fmt = parse_extension(map_file)

        log.debug(f'Initialize map interpolator for file {map_file}')
        map_data = straxen.get_resource(map_file, fmt=fmt)
        return straxen.InterpolatingMap(map_data, method=method)

    else:
        raise TypeError("Can't handle map_file except a string or a list")
@export
def make_patternmap(map_file, fmt=None, method='WeightedNearestNeighbors', pmt_mask=None):
    """ This is special interpretation of the of previous make_map(), but designed
    for pattern map loading with provided PMT mask. This way simplifies both S1 and S2
    cases
    """
    # making tests not failing, we can probably overwrite it completel
    if isinstance(map_file, list):
        log.warning(f'Using dummy map with pattern mask! This has no effect here!')
        assert map_file[0] == 'constant dummy', ('Alternative file input can only be '
                                                 '("constant dummy", constant: int, shape: list')
        return DummyMap(map_file[1], map_file[2])
    elif isinstance(map_file, str):
        if fmt is None:
            fmt = parse_extension(map_file)
        map_data = deepcopy(straxen.get_resource(map_file, fmt=fmt))
        # XXX: straxed deals with pointers and caches resources, it means that resources are global
        # what is bad, so we make own copy here and modify it locally
        if 'compressed' in map_data:
            compressor, dtype, shape = map_data['compressed']
            map_data['map'] = np.frombuffer(
                strax.io.COMPRESSORS[compressor]['decompress'](map_data['map']),
                dtype=dtype).reshape(*shape)
            del map_data['compressed']
        if 'quantized' in map_data:
            map_data['map'] = map_data['quantized']*map_data['map'].astype(np.float32)
            del map_data['quantized']
        if not (pmt_mask is None):
            assert (map_data['map'].shape[-1]==pmt_mask.shape[0]), "Error! Pattern map and PMT gains must have same dimensions!"
            map_data['map'][..., ~pmt_mask]=0.0
        return straxen.InterpolatingMap(map_data, method=method)
    else:
        raise TypeError("Can't handle map_file except a string or a list")

@export
class DummyMap:
    """Return constant results
        the length match the length of input
        but from the second dimensions the shape is user defined input
    """
    def __init__(self, const, shape=()):
        self.const = const
        self.shape = shape

    def __call__(self, x, **kwargs):
        shape = [len(x)] + list(self.shape)
        return np.ones(shape) * self.const

    def reduce_last_dim(self):
        assert len(self.shape) >= 1, 'Need at least 1 dim to reduce further'
        const = self.const * self.shape[-1]
        shape = list(self.shape)
        shape[-1] = 1

        return DummyMap(const, shape)


##
# Copy from https://github.com/XENONnT/private_nt_aux_files
##
def parse_extension(name):
    """Get the extention from a file name. If zipped or tarred, can contain a dot"""
    split_name = name.split('.')
    if len(split_name) == 2:
        fmt = split_name[-1]
    elif len(split_name) > 2 and 'gz' in name:
        fmt = '.'.join(split_name[-2:])
    else:
        fmt = split_name[-1]
    log.warning(f'Using {fmt} for unspecified {name}')
    return fmt



import logging
from numba import njit
import numpy as np
from scipy.stats import skewnorm
from strax import exporter
from .pulse import Pulse
from .. import units
from ..load_resource import DummyMap

export, __all__ = exporter()
logging.basicConfig(handlers=[logging.StreamHandler()])
log = logging.getLogger('wfsim.core')
log.setLevel('WARNING')


@export
class S2(Pulse):
    """
    Given temporal inputs as well as number of electrons
    Random generate photon timing and channel distribution.
    """

    def __init__(self, config):
        super().__init__(config)

        self.phase = 'gas'  # To distinguish singlet/triplet time delay.
        self.luminescence_switch_threshold = 100  # When to use simplified model (NOT IN USE)

    @staticmethod
    def inverse_field_distortion_correction(x, y, z, resource):
        """For 1T the pattern map is a data driven one so we need to reverse engineer field distortion correction
        into the simulated positions
        :param x: 1d array of float
        :param y: 1d array of float
        :param z: 1d array of float
        :param resource: instance of resource class
        returns z: 1d array, postions 2d array 
        """
        positions = np.array([x, y, z]).T
        for i_iter in range(6):  # 6 iterations seems to work
            dr = resource.fdc_3d(positions)
            if i_iter > 0:
                dr = 0.5 * dr + 0.5 * dr_pre  # Average between iter
            dr_pre = dr

            r_obs = np.sqrt(x**2 + y**2) - dr
            x_obs = x * r_obs / (r_obs + dr)
            y_obs = y * r_obs / (r_obs + dr)
            z_obs = - np.sqrt(z**2 + dr**2)
            positions = np.array([x_obs, y_obs, z_obs]).T

        positions = np.array([x_obs, y_obs]).T 
        return z_obs, positions
    
    @staticmethod
    def field_distortion_comsol(x, y, z, resource):
        """Field distortion from the COMSOL simulation for the given electrode configuration:
        :param x: 1d array of float
        :param y: 1d array of float
        :param z: 1d array of float
        :param resource: instance of resource class
        returns z: 1d array, postions 2d array 
        """
        positions = np.array([np.sqrt(x**2 + y**2), z]).T
        theta = np.arctan2(y, x)
        r_obs = resource.fd_comsol(positions, map_name='r_distortion_map')
        x_obs = r_obs * np.cos(theta)
        y_obs = r_obs * np.sin(theta)
        
        positions = np.array([x_obs, y_obs]).T 
        return z, positions

    def __call__(self, instruction):
        if len(instruction.shape) < 1:
            # shape of recarr is a bit strange
            instruction = np.array([instruction])

        _, _, t, x, y, z, n_electron, recoil_type, *rest = [
            np.array(v).reshape(-1) for v in zip(*instruction)]
        
        # Reverse engineering FDC
        if self.config['field_distortion_model'] == 'inverse_fdc':
            z_obs, positions = self.inverse_field_distortion_correction(x, y, z, resource=self.resource)
        # Reverse engineering FDC
        elif self.config['field_distortion_model'] == 'comsol':
            z_obs, positions = self.field_distortion_comsol(x, y, z, resource=self.resource)
        else:
            z_obs, positions = z, np.array([x, y]).T

        n_electron = self.get_electron_yield(n_electron=n_electron,
                                             xy_int= np.array([x, y]).T, # maps are in R_true, so orginal position should be here
                                             z_int=z, # maps are in Z_true, so orginal position should be here
                                             config=self.config,
                                             resource=self.resource)

        sc_gain = self.get_s2_light_yield(positions=positions,
                                          config=self.config,
                                          resource=self.resource)
        


        n_photons_per_xy, n_photons_per_ele, self._electron_timings = self.get_n_photons(t=t,
                                                                                         n_electron=n_electron,
                                                                                         z_int=z,
                                                                                         xy_int= np.array([x, y]).T,
                                                                                         sc_gain=sc_gain, 
                                                                                         config=self.config,
                                                                                         resource=self.resource)

        self._instruction = np.repeat(np.arange(len(n_electron)), n_photons_per_xy)
        self._photon_channels = self.photon_channels(n_electron=n_electron,
                                                     z_obs=z_obs,
                                                     positions=positions,
                                                     _instruction=self._instruction,
                                                     config=self.config,
                                                     resource=self.resource)

        # Second generate photon timing
        self._photon_timings = self.photon_timings(positions=positions, 
                                                   n_photons_per_xy=n_photons_per_xy,
                                                   _electron_timings=self._electron_timings,
                                                   n_photons_per_ele=n_photons_per_ele,
                                                   _photon_channels=self._photon_channels,
                                                   phase=self.phase,
                                                   config=self.config,
                                                   resource=self.resource,)

        # Sorting times according to the channel, as non-explicit sorting
        # is performed later and this breaks timing of individual channels/arrays
        sortind = np.argsort(self._photon_channels)

        self._photon_channels = self._photon_channels[sortind]
        self._photon_timings = self._photon_timings[sortind]

        super().__call__()

    @staticmethod
    def get_avg_drift_velocity(z, xy, config, resource):
        """Calculate s2 drift time mean and spread

        :param z: 1d array of z (floats)
        :param xy: 2d array of xy positions (floats)
        :param config: dict with wfsim config
        :param resource: instance of the resource class

        returns array of floats corresponding to average drift velocities from given point to the gate
        """
        if config['enable_field_dependencies']['drift_speed_map']:
            drift_v_LXe = resource.field_dependencies_map(z, xy, map_name='drift_speed_map')  # mm/µs
            drift_v_LXe *= 1e-4  # cm/ns
            drift_v_LXe *= resource.drift_velocity_scaling
        else:
            drift_v_LXe = config['drift_velocity_liquid']
        return drift_v_LXe

    @staticmethod
    def get_s2_drift_time_params(z_int, xy_int, config, resource):
        """Calculate s2 drift time mean and spread

        :param z_int: 1d array of true z (floats) 
        :param xy_int: 2d array of true xy positions (floats)
        :param config: dict with wfsim config
        :param resource: instance of the resource class

        returns two arrays of floats (mean drift time, drift time spread) 
        """
        drift_velocity_liquid = S2.get_avg_drift_velocity(z_int, xy_int, config, resource)
        if config['enable_field_dependencies']['diffusion_longitudinal_map']:
            diffusion_constant_longitudinal = resource.field_dependencies_map(z_int, xy_int, map_name='diffusion_longitudinal_map')  # cm²/s
            diffusion_constant_longitudinal *= 1e-9  # cm²/ns
        else:
            diffusion_constant_longitudinal = config['diffusion_constant_longitudinal']

        drift_time_mean = - z_int / \
            drift_velocity_liquid + config['drift_time_gate']
        drift_time_mean = np.clip(drift_time_mean, 0, np.inf)
        drift_time_spread = np.sqrt(2 * diffusion_constant_longitudinal * drift_time_mean)
        drift_time_spread /= drift_velocity_liquid
        return drift_time_mean, drift_time_spread
    
    @staticmethod
    def get_s2_light_yield(positions, config, resource):
        """Calculate s2 light yield...
        
        :param positions: 2d array of positions (floats)
        :param config: dict with wfsim config
        :param resource: instance of the resource class
        
        returns array of floats (mean expectation) 
        """
        
        if config.get('se_gain_from_map', False):
           sc_gain = resource.se_gain_map(positions)
        else:
           # calculate it from MC pattern map directly if no "se_gain_map" is given
           sc_gain = resource.s2_correction_map(positions)
           sc_gain *= config['s2_secondary_sc_gain']

       # depending on if you use the data driven or mc pattern map for light yield for S2
       # the shape of n_photon_hits will change. Mc needs a squeeze
        if len(sc_gain.shape) != 1:
            sc_gain=np.squeeze(sc_gain, axis=-1)

        # sc gain should has the unit of pe / electron, here we divide 1 + dpe to get nphoton / electron
        sc_gain /= 1 + config['p_double_pe_emision']

        # data driven map contains nan, will be set to 0 here
        sc_gain[np.isnan(sc_gain)] = 0

        return sc_gain

    @staticmethod
    def get_electron_yield(n_electron, xy_int, z_int, config, resource):
        """Drift electrons up to the gas interface and absorb them

        :param n_electron: 1d array with ints as number of electrons
        :param xy_int: 2d array of xy interaction positions (floats)
        :param z_int: 1d array of floats with the z interaction positions (floats)
        :param config: dict with wfsim config
        :param resource: instance of the resource class

        returns 1d array ints with number of electrons
        """
        # Average drift time of the electrons
        drift_time_mean, drift_time_spread = S2.get_s2_drift_time_params(z_int, xy_int, config, resource)

        # Absorb electrons during the drift
        electron_lifetime_correction = np.exp(- 1 * drift_time_mean /
                                              config['electron_lifetime_liquid'])

        if config.get('ext_eff_from_map', False):
            # Extraction efficiency is g2(x,y)/SE_gain(x,y)
            cy = config['g2_mean']*resource.s2_correction_map(xy_int)* \
                 electron_lifetime_correction/resource.se_gain_map(xy_int)
        else:
            cy = electron_lifetime_correction*config['electron_extraction_yield']

        # Remove electrons in insensitive volume
        if config['enable_field_dependencies']['survival_probability_map']:
            p_surv = resource.field_dependencies_map(z_int, xy_int, map_name='survival_probability_map')
            if np.any(p_surv<0) or np.any(p_surv>1):
                # FIXME: this is necessary due to map artefacts, such as negative or values >1
                p_surv=np.clip(p_surv, a_min = 0, a_max = 1)
            cy *= p_surv
        n_electron = np.random.binomial(n=n_electron, p=cy)
        
        return n_electron

    @staticmethod
    @njit
    def electron_timings(t, n_electron, drift_time_mean, drift_time_spread, sc_gain, timings, gains,
                         electron_trapping_time):
        """Calculate arrival times of the electrons. Data is written to the timings and gains arrays
        :param t: 1d array of ints
        :param n_electron:1 d array of ints
        :param drift_time_mean: 1d array of floats
        :param drift_time_spread: 1d array of floats
        :param sc_gain: secondary scintillation gain       
        :param timings: empty array with length sum(n_electron)
        :param gains: empty array with length sum(n_electron)
        :param electron_trapping_time: configuration values
        """
        assert len(timings) == np.sum(n_electron)
        assert len(gains) == np.sum(n_electron)
        assert len(sc_gain) == len(t)

        i_electron = 0
        for i in np.arange(len(t)):
            # Calculate electron arrival times in the ELR region
            for _ in np.arange(n_electron[i]):
                _timing = np.random.exponential(electron_trapping_time)
                _timing += np.random.normal(drift_time_mean[i], drift_time_spread[i])
                timings[i_electron] = t[i] + int(_timing)

                # add manual fluctuation to sc gain
                gains[i_electron] = sc_gain[i]
                i_electron += 1

    @staticmethod
    def get_n_photons(t, n_electron, z_int, xy_int, sc_gain, config, resource):
        """Generates photon timings for S2s.
        Returns a list of photon timings and instructions repeated for original electron
        
        :param t: 1d int array time of s2
        :param n_electron: 1d float array number of electrons to simulate
        :param z_int: float array. true Z interaction positions
        :param xy_int: 2d float array, true xy interaction positions
        :param sc_gain: float, secondary s2 gain
        :param config: dict of the wfsim config
        :param resource: instance of the resource class """
        # Get electron timings
        drift_time_mean, drift_time_spread = S2.get_s2_drift_time_params(z_int, xy_int, config, resource)
        _electron_timings = np.zeros(np.sum(n_electron), np.int64)
        _electron_gains = np.zeros(np.sum(n_electron), np.float64)
        S2.electron_timings(t, n_electron, drift_time_mean, drift_time_spread, sc_gain,
                            _electron_timings, _electron_gains, config['electron_trapping_time'])

        # Populate with photons per e/ per position
        n_photons_per_ele = np.random.poisson(_electron_gains)
        n_photons_per_ele += np.random.normal(0, config.get('s2_gain_spread', 0), len(n_photons_per_ele)).astype(np.int64)
        n_photons_per_ele[n_photons_per_ele < 0] = 0
        #
        n_photons_per_xy = np.cumsum(np.pad(n_photons_per_ele, [1, 0]))[np.cumsum(n_electron)]
        n_photons_per_xy = np.diff(np.pad(n_photons_per_xy, [1, 0]))

        return n_photons_per_xy, n_photons_per_ele, _electron_timings

    @staticmethod
    @njit
    def _luminescence_timings_simple(n, dG, E0, r, dr, rr, alpha, uE, p, n_photons):
        """
        Luminescence time distribution computation, calculates emission timings of photons from the excited electrons
        return 1d nested array with ints
        """
        emission_time = np.zeros(np.sum(n_photons), np.int64)

        ci = 0
        for i in range(n):
            npho = n_photons[i]
            dt = dr / (alpha * E0[i] * rr)
            dy = E0[i] * rr / uE - 0.8 * p  # arXiv:physics/0702142
            avgt = np.sum(np.cumsum(dt) * dy) / np.sum(dy)

            j = np.argmax(r <= dG[i])
            t = np.cumsum(dt[j:]) - avgt
            y = np.cumsum(dy[j:])

            probabilities = np.random.rand(npho)
            emission_time[ci:ci+npho] = np.interp(probabilities, y / y[-1], t).astype(np.int64)
            ci += npho

        return emission_time

    @staticmethod
    def luminescence_timings_simple(xy, n_photons, config, resource):
        """
        Luminescence time distribution computation according to simple s2 model (many many many single electrons)
        :param xy: 1d array with positions
        :param n_photons: 1d array with ints for number of xy positions
        :param config: dict wfsim config
        :param resource: instance of wfsim resource
        returns _luminescence_timings_simple
        """
        assert len(n_photons) == len(xy), 'Input number of n_photons should have same length as positions'

        number_density_gas = config['pressure'] / \
            (units.boltzmannConstant * config['temperature'])
        alpha = config['gas_drift_velocity_slope'] / number_density_gas
        uE = units.kV / units.cm
        pressure = config['pressure'] / units.bar

        if config.get('enable_gas_gap_warping', True):
            dG = resource.gas_gap_length(xy)
        else:
            dG = np.ones(len(xy)) * config['elr_gas_gap_length']
        rA = config['anode_field_domination_distance']
        rW = config['anode_wire_radius']
        dL = config['gate_to_anode_distance'] - dG

        VG = config['anode_voltage'] / (1 + dL / dG / config['lxe_dielectric_constant'])
        E0 = VG / ((dG - rA) / rA + np.log(rA / rW))  # V / cm

        dr = 0.0001  # cm
        r = np.arange(np.max(dG), rW, -dr)
        rr = np.clip(1 / r, 1 / rA, 1 / rW)

        return S2._luminescence_timings_simple(len(xy), dG, E0, 
                                               r, dr, rr, alpha, uE,
                                               pressure, n_photons)
    
    @staticmethod
    def luminescence_timings_garfield(xy, n_photons, config, resource, confine_position=None):
        """
        Luminescence time distribution computation according to garfield scintillation maps
        :param xy: 1d array with positions
        :param n_photons: 1d array with ints for number of xy positions
        :param config: dict wfsim config
        :param resource: instance of wfsim resource
        :param confine_position: if float, confine extraction region +/- this position around anode wires
        returns 2d array with ints for photon timings of input param 'shape'
        """
        assert 's2_luminescence' in resource.__dict__, 's2_luminescence model not found'
        assert len(n_photons) == len(xy), 'Input number of n_electron should have same length as positions'
        assert len(resource.s2_luminescence['t'].shape) == 2, 'Timing data is expected to have D2'

        if type(confine_position)==float:
            distance = np.random.uniform(-confine_position, confine_position, len(xy))
        else:
            tilt = config.get('anode_xaxis_angle', np.pi / 4)
            pitch = config.get('anode_pitch', 0.5)
            rotation_mat = np.array(((np.cos(tilt), -np.sin(tilt)), (np.sin(tilt), np.cos(tilt))))
            jagged = lambda relative_y: (relative_y + pitch / 2) % pitch - pitch / 2
            distance = jagged(np.matmul(xy, rotation_mat)[:, 1])  # shortest distance from any wire

        index_row = [np.argmin(np.abs(d - resource.s2_luminescence['x'])) for d in distance]
        index_row = np.repeat(index_row, n_photons).astype(np.int64)
        index_col = np.random.randint(0, resource.s2_luminescence['t'].shape[1], np.sum(n_photons), np.int64)
        
        avgt = np.average(resource.s2_luminescence['t']).astype(int)
        return resource.s2_luminescence['t'][index_row, index_col].astype(np.int64) - avgt
    
    @staticmethod
    @njit
    def draw_excitation_times(inv_cdf_list, hist_indices, nph, diff_nearest_gg, d_gas_gap):
        
        """
        Draws the excitation times from the GARFIELD electroluminescence map
        
        :param inv_cdf_list:       List of inverse CDFs for the excitation time histograms
        :param hist_indices:       The index of the histogram which refers to the gas gap
        :param nph:                A 1-d array of the number of photons per electron
        :param diff_nearest_gg:    The difference between the gas gap from the map
                                   (continuous value) and the nearest (discrete) value
                                   of the gas gap corresponding to the excitation time
                                   histograms
        :param d_gas_gap:          Spacing between two consecutive gas gap values
        
        returns time of each photon
        """
        
        inv_cdf_len = len(inv_cdf_list[0])
        timings = np.zeros(np.sum(nph))
        upper_hist_ind = np.clip(hist_indices+1, 0, len(inv_cdf_list)-1)
        
        count = 0
        for i, (hist_ind, u_hist_ind, n, dngg) in enumerate(zip(hist_indices,
                                                                upper_hist_ind, 
                                                                nph,
                                                                diff_nearest_gg)):
            
            #There are only 10 values of gas gap separated by 0.1mm, so we interpolate
            #between two histograms
            
            interp_cdf = ((inv_cdf_list[u_hist_ind]-inv_cdf_list[hist_ind])*(dngg/d_gas_gap)
                           +inv_cdf_list[hist_ind])

            samples = np.random.uniform(0, inv_cdf_len-1, n)
            t1 = interp_cdf[np.floor(samples).astype('int')]
            t2 = interp_cdf[np.ceil(samples).astype('int')]
            T = (t2-t1)*(samples - np.floor(samples))+t1
            if n!=0:
                T = T-np.mean(T)
            
            #subtract mean to get proper drift time and z correlation
            timings[count:count+n] = T
            count+=n
        return timings
    
    @staticmethod
    def luminescence_timings_garfield_gasgap(xy, n_photons, resource):
        """
        Luminescence time distribution computation according to garfield scintillation maps
        which are ONLY drawn from below the anode, and at different gas gaps
        :param xy: 1d array with positions
        :param n_photons: 1d array with ints for number of xy positions
        :param resource: instance of wfsim resource

        returns 2d array with ints for photon timings of input param 'shape'
        """
        assert 's2_luminescence_gg' in resource.__dict__, 's2_luminescence_gg model not found'
        assert len(n_photons) == len(xy), 'Input number of n_electron should have same length as positions'
        
        d_gasgap = resource.s2_luminescence_gg['gas_gap'][1]-resource.s2_luminescence_gg['gas_gap'][0]
        
        cont_gas_gaps = resource.garfield_gas_gap_map(xy)
        draw_index = np.digitize(cont_gas_gaps, resource.s2_luminescence_gg['gas_gap'])-1
        diff_nearest_gg = cont_gas_gaps - resource.s2_luminescence_gg['gas_gap'][draw_index]
        
        return S2.draw_excitation_times(resource.s2_luminescence_gg['timing_inv_cdf'],
                                        draw_index,
                                        n_photons,
                                        diff_nearest_gg,
                                        d_gasgap)
    
    @staticmethod
    def optical_propagation(channels, config, spline):
        """Function getting times from s2 timing splines:
        :param channels: The channels of all s2 photon
        :param config: current configuration of wfsim
        :param spline: pointer to s2 optical propagation splines from resources
        """
        prop_time = np.zeros_like(channels)
        u_rand = np.random.rand(len(channels))[:, None]

        is_top = channels < config['n_top_pmts']
        prop_time[is_top] = spline(u_rand[is_top], map_name='top')

        is_bottom = channels >= config['n_top_pmts']
        prop_time[is_bottom] = spline(u_rand[is_bottom], map_name='bottom')

        return prop_time.astype(np.int64)

    @staticmethod
    def photon_timings(positions, n_photons_per_xy, _electron_timings, n_photons_per_ele, _photon_channels, phase, config, resource):
        """Get photon times and add delays based on models

        :param positions: 2d float array, xy positions of s2
        :param n_photons_per_xy: number of photons for each xy position
        :param _electron_timings: electron timings
        :param n_photons_per_ele: number of photons for each electron
        :param _photon_channels: channel of each photon
        :param config: dict of the wfsim config
        :param phase: gas
        :param resource: instance of the resource class
        """

        # Luminescence Timings
        if config['s2_luminescence_model']=='simple':
            _photon_timings = S2.luminescence_timings_simple(positions, n_photons_per_xy,
                                                             config=config,
                                                             resource=resource)
        elif config['s2_luminescence_model']=='garfield':
            confine_position=None
            if 's2_garfield_confine_position' in config:
                if config['s2_garfield_confine_position'] > 0.0:
                    confine_position=config['s2_garfield_confine_position']
            _photon_timings = S2.luminescence_timings_garfield(positions, n_photons_per_xy,
                                                               config=config,
                                                               resource=resource,
                                                               confine_position=confine_position)
            
        elif config['s2_luminescence_model']=='garfield_gas_gap':
            _photon_timings = S2.luminescence_timings_garfield_gasgap(positions, n_photons_per_xy,
                                                                      resource=resource)
        else:
            raise KeyError(f"{config['s2_luminescence_model']} is not valid! Use 'simple' or 'garfield' or 'garfield_gas_gap'")

        # Emission Delay
        _photon_timings += Pulse.singlet_triplet_delays(len(_photon_timings), config['singlet_fraction_gas'], config, phase)

        # Optical Propagation Delay
        if "optical_propagation" in config['s2_time_model']:
            # optical propagation splitting top and bottom
            _photon_timings += S2.optical_propagation(_photon_channels, config, resource.s2_optical_propagation_spline)
        elif "zero_delay" in config['s2_time_model']:
            # no optical propagation delay
            _photon_timings += np.zeros_like(_photon_timings, dtype=np.int64)
        elif "s2_time_spread around zero" in config['s2_time_model']:
            # simple/existing delay
            _photon_timings += np.random.normal(0, config['s2_time_spread'], len(_photon_timings)).astype(np.int64)
        else:
            raise KeyError(f"{config['s2_time_model']} is not in any of the valid s2 time models")

        # repeat for n photons per electron # Should this be before adding delays?
        _photon_timings += np.repeat(_electron_timings, n_photons_per_ele)

        return _photon_timings

    @staticmethod
    def s2_pattern_map_diffuse(n_electron, z, xy, config, resource):
        """Returns an array of pattern of shape [n interaction, n PMTs]
        pattern of each interaction is an average of n_electron patterns evaluated at
        diffused position near xy. The diffused positions sample from 2d symmetric gaussian
        with spread scale with sqrt of drift time.

        :param n_electron: a 1d int array
        :param z: a 1d float array
        :param xy: a 2d float array of shape [n interaction, 2]
        :param config: dict of the wfsim config
        :param resource: instance of the resource class
        """
        assert all(z < 0), 'All S2 in liquid should have z < 0'
        drift_velocity_liquid=S2.get_avg_drift_velocity(z, xy, config, resource)

        if config['enable_field_dependencies']['diffusion_transverse_map']:
            diffusion_constant_radial = resource.field_dependencies_map(z, xy, map_name='diffusion_radial_map')  # cm²/s
            diffusion_constant_azimuthal = resource.field_dependencies_map(z, xy, map_name='diffusion_azimuthal_map') # cm²/s
            diffusion_constant_radial *= 1e-9  # cm²/ns
            diffusion_constant_azimuthal *= 1e-9  # cm²/ns
        else:
            diffusion_constant_transverse = getattr(config, 'diffusion_constant_transverse', 0)
            diffusion_constant_radial = diffusion_constant_transverse
            diffusion_constant_azimuthal = diffusion_constant_transverse

        drift_time_mean = - z / drift_velocity_liquid
        hdiff_stdev_radial = np.sqrt(2 * diffusion_constant_radial * drift_time_mean)
        hdiff_stdev_azimuthal = np.sqrt(2 * diffusion_constant_azimuthal * drift_time_mean)
        
        hdiff_radial = np.random.normal(0, 1, np.sum(n_electron)) * np.repeat(hdiff_stdev_radial, n_electron, axis=0)
        hdiff_azimuthal = np.random.normal(0, 1, np.sum(n_electron)) * np.repeat(hdiff_stdev_azimuthal, n_electron, axis=0)
        hdiff = np.column_stack([hdiff_radial, hdiff_azimuthal])
        theta = np.arctan2(xy[:,1], xy[:,0])
        matrix = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]).T
        hdiff = np.vstack([(matrix[i] @ np.split(hdiff, np.cumsum(n_electron))[:-1][i].T).T for i in range(len(matrix))])
        # Should we also output this xy position in truth?
        xy_multi = np.repeat(xy, n_electron, axis=0) + hdiff  # One entry xy per electron
        # Remove points outside tpc, and the pattern will be the average inside tpc
        # Should be done naturally with the s2 pattern map, however, there's some bug there, so we apply this hard cut
        mask = np.sum(xy_multi ** 2, axis=1) <= config['tpc_radius'] ** 2

        if isinstance(resource.s2_pattern_map, DummyMap):
            output_dim = resource.s2_pattern_map.shape[-1]
        else:
            output_dim = resource.s2_pattern_map.data['map'].shape[-1]
        pattern = np.zeros((len(n_electron), output_dim))
        n0 = 0
        # Average over electrons for each s2
        for ix, ne in enumerate(n_electron):
            s = slice(n0, n0+ne)
            pattern[ix, :] = np.average(resource.s2_pattern_map(xy_multi[s][mask[s]]), axis=0)
            n0 += ne

        return pattern

    @staticmethod
    def photon_channels(n_electron, z_obs, positions, _instruction, config, resource):
        """Set the _photon_channels property list of length same as _photon_timings
        :param n_electron: a 1d int array
        :param z_obs: a 1d float array
        :param positions: a 2d float array of shape [n interaction, 2] for the xy coordinate
        :param _instruction: array of instructions with dtype wfsim.instructions_dtype
        :param config: dict wfsim config
        :param resource: instance of resource class
        """
        if len(_instruction) == 0:
            _photon_channels = np.zeros(0, dtype=np.int64)
            return _photon_channels

        #aft = config['s2_mean_area_fraction_top']
        aft_sigma = config.get('s2_aft_sigma', 0.0)
        aft_skewness = config.get('s2_aft_skewness', 0.0)

        channels = np.arange(config['n_tpc_pmts']).astype(np.int64)
        top_index = np.arange(config['n_top_pmts'])
        bottom_index = np.array(config['channels_bottom'])

        if config.get('diffusion_constant_transverse', 0) > 0:
            pattern = S2.s2_pattern_map_diffuse(n_electron, z_obs, positions, config, resource)  # [position, pmt]
        else:
            pattern = resource.s2_pattern_map(positions)  # [position, pmt]

        if pattern.shape[1] - 1 not in bottom_index:
            pattern = np.pad(pattern, [[0, 0], [0, len(bottom_index)]], 
                             'constant', constant_values=1)

        # Remove turned off pmts
        pattern[:, np.in1d(channels, config['turned_off_pmts'])] = 0

        sum_pat = np.sum(pattern, axis=1).reshape(-1, 1)
        pattern = np.divide(pattern, sum_pat, out=np.zeros_like(pattern), where=sum_pat != 0)

        assert pattern.shape[0] == len(positions)
        assert pattern.shape[1] == len(channels)

        _buffer_photon_channels = []
        # Randomly assign to channel given probability of each channel
        for unique_i, count in zip(*np.unique(_instruction, return_counts=True)):
            pat = pattern[unique_i]  # [pmt]

            if aft_sigma != 0:  # Redistribute pattern with user specified aft smearing
                _cur_aft=np.sum(pat[top_index])/np.sum(pat)
                _new_aft=_cur_aft*skewnorm.rvs(loc=1.0, scale=aft_sigma, a=aft_skewness)
                _new_aft=np.clip(_new_aft, 0, 1)
                pat[top_index]*=(_new_aft/_cur_aft)
                pat[bottom_index]*=(1 - _new_aft)/(1 - _cur_aft)
                # old code from Peter
                #pat[top_index] = pat[top_index] / pat[top_index].sum() * _aft
                #pat[bottom_index] = pat[bottom_index] / pat[bottom_index].sum() * (1 - _aft)

            if np.isnan(pat).sum() > 0:  # Pattern map return zeros
                _photon_channels = np.array([-1] * count)
            else:
                _photon_channels = np.random.choice(
                    channels,
                    size=count,
                    p=pat,
                    replace=True)

            _buffer_photon_channels.append(_photon_channels)

        _photon_channels = np.concatenate(_buffer_photon_channels)
        return _photon_channels



__version__ = "0.7.0"

from .core.afterpulse import *
from .core.pulse import *
from .core.rawdata import *
from .core.s1 import *
from .core.s2 import *

from .pax_interface import *
from .strax_interface import *

from .load_resource import *
from .utils import *



import logging
from numba import njit
import numpy as np
from scipy.interpolate import interp1d
from strax import exporter, deterministic_hash
from ..load_resource import load_config


export, __all__ = exporter()
logging.basicConfig(handlers=[logging.StreamHandler()])
log = logging.getLogger('wfsim.core')
log.setLevel('WARNING')


__all__ += ['_cached_pmt_current_templates', '_cached_uniform_to_pe_arr']
_cached_pmt_current_templates = {}
_cached_uniform_to_pe_arr = {}


@export
class Pulse(object):
    """Pulse building class"""

    def __init__(self, config):
        self.config = config
        self.config.update(self.config.get(self.__class__.__name__, {}))
        self.resource = load_config(config)

        self.init_pmt_current_templates()
        self.init_spe_scaling_factor_distributions()
        self.config['turned_off_pmts'] = np.arange(len(config['gains']))[np.array(config['gains']) == 0]
        self.current_max = np.max(np.array(self._pmt_current_templates), axis=1)
        self.current_2_adc = self.config['pmt_circuit_load_resistor'] \
            * self.config['external_amplification'] \
            / (self.config['digitizer_voltage_range'] / 2 ** (self.config['digitizer_bits']))

        self.clear_pulse_cache()

    def __call__(self, *args):
        """
        PMTs' response to incident photons
        Use _photon_timings, _photon_channels to build pulses
        """
        if ('_photon_timings' not in self.__dict__) or \
                ('_photon_channels' not in self.__dict__):
            raise NotImplementedError

        # The pulse cache should be immediately transferred after call this function
        self.clear_pulse_cache()

        # Correct for PMT Transition Time Spread (skip for pmt after-pulses)
        # note that PMT datasheet provides FWHM TTS, so sigma = TTS/(2*sqrt(2*log(2)))=TTS/2.35482
        if '_photon_gains' not in self.__dict__:
            self._photon_timings += np.random.normal(self.config['pmt_transit_time_mean'],
                                                     self.config['pmt_transit_time_spread'] / 2.35482,
                                                     len(self._photon_timings)).astype(np.int64)

        dt = self.config.get('sample_duration', 10)  # Getting dt from the lib just once
        self._n_photon = self._n_photon_bottom = 0  # For truth output
        self._n_pe = self._n_pe_bottom = 0
        self._n_photon_trigger = self._n_photon_trigger_bottom = 0
        self._n_pe_trigger = self._n_pe_trigger_bottom = 0
        self._raw_area = self._raw_area_bottom = 0
        self._raw_area_trigger = self._raw_area_trigger_bottom = 0

        # Assign DPE, and save it for compute after-pulses
        p_dpe = self.config['p_double_pe_emision']
        self._photon_is_dpe = np.random.binomial(n=1,
                                                p=p_dpe,
                                                size=len(self._photon_timings)).astype(np.bool_)

        counts_start = 0  # Secondary loop index for assigning channel
        for channel, counts in zip(*np.unique(self._photon_channels, return_counts=True)):

            # Use 'counts' amount of photon for this channel
            _channel_photon_timings = self._photon_timings[counts_start:counts_start+counts]
            _channel_photon_is_dpe = self._photon_is_dpe[counts_start:counts_start+counts]

            counts_start += counts
            if channel in self.config['turned_off_pmts']:
                continue

            # If gain of each photon is not specifically assigned
            # Sample from spe scaling factor distribution and to individual gain
            # In contrast to pmt afterpulse that should have gain determined before this step
            if '_photon_gains' not in self.__dict__:

                _channel_photon_gains = self.config['gains'][channel] \
                    * self.uniform_to_pe_arr(np.random.random(len(_channel_photon_timings)), channel)

                # Add some double photoelectron emission by adding another sampled gain
                n_double_pe = _channel_photon_is_dpe.sum()
                _channel_photon_gains[_channel_photon_is_dpe] += self.config['gains'][channel] \
                    * self.uniform_to_pe_arr(np.random.random(n_double_pe), channel)

            else:
                n_double_pe = 0
                _channel_photon_gains = np.array(self._photon_gains[self._photon_channels == channel])

            # += truth per channel to internal truth fields
            self.add_truth(
                _channel_photon_timings,
                _channel_photon_gains,
                dt,
                channel,
                n_double_pe,)

            # Build a simulated waveform, length depends on min and max of photon timings
            min_timing, max_timing = np.min(
                _channel_photon_timings), np.max(_channel_photon_timings)

            pulse_left = (int(min_timing // dt) 
                          - int(self.config['samples_to_store_before'])
                          - self.config.get('samples_before_pulse_center', 2))

            pulse_right = (int(max_timing // dt) 
                           + int(self.config['samples_to_store_after'])
                           + self.config.get('samples_after_pulse_center', 20))
            pulse_current = np.zeros(pulse_right - pulse_left + 1)

            Pulse.add_current(_channel_photon_timings.astype(np.int64),
                              _channel_photon_gains,
                              pulse_left,
                              dt,
                              self._pmt_current_templates,
                              pulse_current)

            # For single event, data of pulse level is small enough to store in dataframe
            self._pulses.append(dict(
                photons  = len(_channel_photon_timings),
                channel  = channel,
                left     = pulse_left,
                right    = pulse_right,
                duration = pulse_right - pulse_left + 1,
                current  = pulse_current,))

    def init_pmt_current_templates(self):
        """
        Create spe templates, for 10ns sample duration and 1ns rounding we have:
        _pmt_current_templates[i] : photon timing fall between [10*m+i, 10*m+i+1)
        (i, m are integers)
        """
        h = deterministic_hash(self.config)
        if h in _cached_pmt_current_templates:
            self._pmt_current_templates = _cached_pmt_current_templates[h]
            return

        # Interpolate on cdf ensures that each spe pulse would sum up to 1 pe*sample duration^-1
        pe_pulse_function = interp1d(
            self.config.get('pe_pulse_ts'),
            np.cumsum(self.config.get('pe_pulse_ys')),
            bounds_error=False, fill_value=(0, 1))

        # Samples are always multiples of sample_duration
        sample_duration = self.config.get('sample_duration', 10)
        samples_before = self.config.get('samples_before_pulse_center', 2)
        samples_after = self.config.get('samples_after_pulse_center', 20)
        pmt_pulse_time_rounding = self.config.get('pmt_pulse_time_rounding', 1.0)

        # Let's fix this, so everything can be turned into int
        assert pmt_pulse_time_rounding == 1

        samples = np.linspace(-samples_before * sample_duration,
                              + samples_after * sample_duration,
                              1 + samples_before + samples_after)
        self._template_length = np.int64(len(samples) - 1)

        templates = []
        for r in np.arange(0, sample_duration, pmt_pulse_time_rounding):
            pmt_current = np.diff(pe_pulse_function(samples - r)) / sample_duration  # pe / 10 ns
            # Normalize here to counter tiny rounding error from interpolation
            pmt_current *= (1 / sample_duration) / np.sum(pmt_current)  # pe / 10 ns
            templates.append(pmt_current)
        self._pmt_current_templates = np.array(templates)
        _cached_pmt_current_templates[h] = self._pmt_current_templates

        log.debug('Spe waveform templates created with %s ns resolution, cached with key %s'
                  % (pmt_pulse_time_rounding, h))

    def init_spe_scaling_factor_distributions(self):
        h = deterministic_hash(self.config)
        if h in _cached_uniform_to_pe_arr:
            self.__uniform_to_pe_arr = _cached_uniform_to_pe_arr[h]
            return

        # Extract the spe pdf from a csv file into a pandas dataframe
        spe_shapes = self.resource.photon_area_distribution

        # Create a converter array from uniform random numbers to SPE gains (one interpolator per channel)
        # Scale the distributions so that they have an SPE mean of 1 and then calculate the cdf
        uniform_to_pe_arr = []
        for ch in spe_shapes.columns[1:]:  # skip the first element which is the 'charge' header
            if spe_shapes[ch].sum() > 0:
                # mean_spe = (spe_shapes['charge'].values * spe_shapes[ch]).sum() / spe_shapes[ch].sum()
                scaled_bins = spe_shapes['charge'].values  # / mean_spe
                cdf = np.cumsum(spe_shapes[ch]) / np.sum(spe_shapes[ch])
            else:
                # if sum is 0, just make some dummy axes to pass to interpolator
                cdf = np.linspace(0, 1, 10)
                scaled_bins = np.zeros_like(cdf)

            grid_cdf = np.linspace(0, 1, 2001)
            grid_scale = interp1d(cdf, scaled_bins,
                                  kind='next',
                                  bounds_error=False,
                                  fill_value=(scaled_bins[0], scaled_bins[-1]))(grid_cdf)

            uniform_to_pe_arr.append(grid_scale)

        if len(uniform_to_pe_arr):
            self.__uniform_to_pe_arr = np.stack(uniform_to_pe_arr)
            _cached_uniform_to_pe_arr[h] = self.__uniform_to_pe_arr

        log.debug('Spe scaling factors created, cached with key %s' % h)

    def uniform_to_pe_arr(self, p, channel=0):
        indices = (p * 2000).astype(np.int64) + 1
        return self.__uniform_to_pe_arr[channel, indices]

    def add_truth(self,
                  photon_timings,
                  photon_gains,
                  dt,
                  channel,
                  n_double_pe,
                  ):
        """Add required information to the fields used for the truth information"""
        if channel in self.config['turned_off_pmts']:
            return

        if str(channel) in self.config.get('special_thresholds', {}):
            threshold = self.config['special_thresholds'][str(channel)] - 0.5
        else:
            threshold = self.config['zle_threshold'] - 0.5
        # Figure out if we were above threshold
        # - Current max is the highest value in the SPE pulse model given a
        #   remainder [0-9] ns (offset from 10 ns sample time).
        #   The SPE pulse model sums up to 0.1 (pe/ns), and the peak is ~0.03 (pe/ns)
        # - Multiply this by the sampled gain of each photon, we get (electron/ns)
        # - The current_2_adc take into account n_electron -> current -> voltage
        #   -> amplification -> digitization, converting the electron/ns to adc value.
        remainder = (photon_timings % dt).astype(int)
        max_amplitude_adc = photon_gains * self.current_max[remainder] * self.current_2_adc
        above_threshold = max_amplitude_adc > threshold
        trigger_photon = np.sum(above_threshold)
        trigger_dpe = np.sum(above_threshold[:n_double_pe])
        raw_area = np.sum(photon_gains) / self.config['gains'][channel]
        trigger_raw_area = np.sum(photon_gains[above_threshold]) / self.config['gains'][channel]

        self._n_photon += len(photon_timings)
        self._n_pe += len(photon_timings) + n_double_pe
        self._n_photon_trigger += trigger_photon
        self._n_pe_trigger += trigger_photon + trigger_dpe
        self._raw_area += raw_area
        self._raw_area_trigger += trigger_raw_area

        if channel in self.config['channels_bottom']:
            self._n_photon_bottom += len(photon_timings)
            self._n_pe_bottom += len(photon_timings) + n_double_pe
            self._n_photon_trigger_bottom += trigger_photon
            self._n_pe_trigger_bottom += trigger_photon + trigger_dpe
            self._raw_area_bottom += raw_area
            self._raw_area_trigger_bottom += trigger_raw_area

    def clear_pulse_cache(self):
        self._pulses = []

    @staticmethod
    @njit
    def add_current(photon_timings,
                    photon_gains,
                    pulse_left,
                    dt,
                    pmt_current_templates,
                    pulse_current):
        #         """
        #         Simulate single channel waveform given the photon timings
        #         photon_timing         - dim-1 integer array of photon timings in unit of ns
        #         photon_gain           - dim-1 float array of ph. 2 el. gain individual photons
        #         pulse_left            - left of the pulse in unit of 10 ns
        #         dt                    - mostly it is 10 ns
        #         pmt_current_templates - list of spe templates of different reminders
        #         pulse_current         - waveform
        #         """
        if not len(photon_timings):
            return
        
        template_length = len(pmt_current_templates[0])
        i_photons = np.argsort(photon_timings)
        # Convert photon_timings to int outside this function
        # photon_timings = photon_timings // 1

        gain_total = 0
        tmp_photon_timing = photon_timings[i_photons[0]]
        for i in i_photons:
            if photon_timings[i] > tmp_photon_timing:
                start = int(tmp_photon_timing // dt) - pulse_left
                reminder = int(tmp_photon_timing % dt)
                pulse_current[start:start + template_length] += \
                    pmt_current_templates[reminder] * gain_total

                gain_total = photon_gains[i]
                tmp_photon_timing = photon_timings[i]
            else:
                gain_total += photon_gains[i]

        start = int(tmp_photon_timing // dt) - pulse_left
        reminder = int(tmp_photon_timing % dt)
        pulse_current[start:start + template_length] += \
            pmt_current_templates[reminder] * gain_total

    @staticmethod
    def singlet_triplet_delays(size, singlet_ratio, config, phase):
        """
        Given the amount of the excimer, return time between excimer decay
        and their time of generation.
        size           - amount of excimer
        self.phase     - 'liquid' or 'gas'
        singlet_ratio  - fraction of excimers that become singlets
                         (NOT the ratio of singlets/triplets!)
        """
        if phase == 'liquid':
            t1, t3 = (config['singlet_lifetime_liquid'],
                      config['triplet_lifetime_liquid'])
        elif phase == 'gas':
            t1, t3 = (config['singlet_lifetime_gas'],
                      config['triplet_lifetime_gas'])
        else:
            t1, t3 = 0, 0

        delay = np.random.choice([t1, t3], size, replace=True,
                                 p=[singlet_ratio, 1 - singlet_ratio])
        return (np.random.exponential(1, size) * delay).astype(np.int64)
